module;

#include <Windows.h>
#include <winternl.h>
export module pige.gdi;
export import pige.winapi;
import std;

namespace pige::gdi
{
#pragma warning(push)
#pragma warning(disable: 4201)
	struct key_flags
	{
		constexpr key_flags(LPARAM lparam) noexcept
			: flags{ static_cast<DWORD>(lparam) }
		{ }

		constexpr key_flags() noexcept
			: flags{}
		{ };

		union
		{
			DWORD flags;
			struct
			{
				USHORT repeat_count;
				BYTE scan_code;
				bool is_extended : 5;
				bool context : 1;
				bool previous_state : 1;
				bool transition : 1;
			};
		};
	};
	
#define GDI_OBJECT_TYPES \
	X(none,		0x00) \
	X(context,	0x01) \
	X(region,	0x04) \
	X(bitmap,	0x05) \
	X(palette,	0x08) \
	X(font,		0x0a) \
	X(brush,	0x10) \
	X(emeta,	0x21) \
	X(pen,		0x30) \
	X(epen,		0x50)

	enum class object_type : BYTE
	{
#define X(id, val) id = val,
		GDI_OBJECT_TYPES
#undef X
	};

	LPCWSTR object_name(object_type type)
	{
		switch (type)
		{
#define X(id, val) case object_type::id: return L ## #id;
			GDI_OBJECT_TYPES
#undef X
			default:
				return nullptr;
		}
	}

	struct gdi_handle
	{
		gdi_handle(HGDIOBJ h)
			: handle{ h }
		{ }

		gdi_handle(HDC h)
			: gdi_handle{ reinterpret_cast<HGDIOBJ>(h) }
		{ }

		gdi_handle(HBITMAP h)
			: gdi_handle{ reinterpret_cast<HGDIOBJ>(h) }
		{ }

		gdi_handle(HPALETTE h)
			: gdi_handle{ reinterpret_cast<HGDIOBJ>(h) }
		{ }

		gdi_handle(HFONT h)
			: gdi_handle{ reinterpret_cast<HGDIOBJ>(h) }
		{ }

		gdi_handle(HBRUSH h)
			: gdi_handle{ reinterpret_cast<HGDIOBJ>(h) }
		{ }

		gdi_handle(HPEN h)
			: gdi_handle{ reinterpret_cast<HGDIOBJ>(h) }
		{ }

		union
		{
			HGDIOBJ handle;
			struct
			{
				USHORT index;
				object_type type : 7;
				bool is_stock : 1;
				BYTE generation;
			};
		};

		[[nodiscard]] inline operator HGDIOBJ() const noexcept { return handle; }
	};
#pragma warning(pop)
	/*
	

struct GDICELL
{
	void* pKernelAddress;
	USHORT wProcessId;
	USHORT wCount;
	USHORT wUpper;
	USHORT wType;
	void* pUserAddress;
};
	*/
	struct object_cell
	{
		void *kernel_ptr;
		USHORT process_id;
		USHORT count; //unused?
		object_type type : 7;
		bool is_stock : 1;
		BYTE generation;
		object_type type_internal;
		BYTE type_variant;
		void *user_ptr;

		[[nodiscard]] HGDIOBJ get_handle(USHORT index) const noexcept
		{
			auto handle = ((is_stock << 7 | std::to_underlying(type)) << 16 | index);
			return reinterpret_cast<HGDIOBJ>(static_cast<INT_PTR>(handle)); //sign-extend and cast to pointer
		}
	};

	class object_table
	{
		object_cell *m_table;

	public:
		object_table()
			: m_table{ GetGDITable() }
		{ }

		void list_objects()
		{
			int max_i = 0;
			int mine = 0;
			int system = 0;
			int other = 0;
			int empty = 0;
			auto my_process = GetCurrentProcessId();
			OutputDebugStringW(L"GDI Object Table\nidx   proc stock type    gen i_type  var  count pkernel            puser\n");
			for (auto i = 0u; i <= std::numeric_limits<USHORT>::max(); i++)
			{
				auto const cell = m_table[i];
				//go if empty process and not stock  
				//go if non-empty process other than mine 
				//(!proc && !stock) || (proc && !(proc==mine))
				//if (cell.is_stock && cell.type == object_type::bitmap || !cell.is_stock && cell.process_id != my_process)
				if (cell.process_id != my_process)
				{
					// It happens
					if (cell.type == object_type::none)
						++empty;
					else
					{
						max_i = i;
						if (cell.process_id)
							++other;
						else
							++system;
					}
					continue;
				}
				gdi_handle gdiHandle = cell.get_handle(static_cast<USHORT>(i));
				
				if (cell.type != object_type::none)
				{
					if (cell.process_id)
						++mine;
					else
						++system;
					max_i = i;
					auto name = object_name(cell.type);
					auto iname = object_name(cell.type_internal);
					OutputDebugStringW(std::format(L"{:05} {:04x} {:5} {:7} {:03} {:7} {:#04x} {:05} {:#018x} {:#018x}[{:#018x}]\n",
						i,
						cell.process_id,
						cell.is_stock ? L"stock" : L"",
						name ? name : std::format(L"unk{:#04x}", std::to_underlying(cell.type)).c_str(),
						cell.generation,
						iname ? name : std::format(L"unk{:#04x}", std::to_underlying(cell.type_internal)).c_str(),
						cell.type_variant,
						cell.count,
						reinterpret_cast<UINT_PTR>(cell.kernel_ptr),
						reinterpret_cast<UINT_PTR>(cell.user_ptr),
						cell.user_ptr ? *reinterpret_cast<UINT_PTR *>(cell.user_ptr) : 0
					).c_str());
				}
				else
					++empty;
			}
			OutputDebugStringW(std::format(L"Total: {}, Mine: {}, System: {}, Other: {}; Max Idx: {}\n", mine+system+other, mine, system, other, max_i).c_str());
		}

	private:
		static object_cell * GetGDITable()
		{
			PROCESS_BASIC_INFORMATION processBasicInformation;
			DWORD returnLength = 0;
			NtQueryInformationProcess(
				GetCurrentProcess(),
				ProcessBasicInformation,
				&processBasicInformation,
				sizeof(processBasicInformation),
				&returnLength);

			auto const gdiTable = reinterpret_cast<object_cell *>(processBasicInformation.PebBaseAddress->Reserved9[18]/*unfortunately, the field is not documented*/);

			return gdiTable;
		}

	};

	export class gdi_window : public pige::winapi::window
	{
	public:
		using window::window;

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_KEYDOWN:
					on_keydown(LOWORD(wparam), lparam);
					return 0;
				case WM_PAINT:
					on_paint();
					return 0;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		void on_paint()
		{
			PAINTSTRUCT ps{ };
			auto hdc = BeginPaint(*this, &ps);
			SetBkColor(hdc, RGB(255, 31, 128));
			auto old_br = SelectObject(hdc, CreateHatchBrush(HS_HORIZONTAL, RGB(0, 224, 128)));
			PatBlt(hdc, 0, 0, 100, 100, PATCOPY);
			DeleteObject(SelectObject(hdc, old_br));
			int modes[] = { STRETCH_ANDSCANS, STRETCH_ORSCANS, STRETCH_DELETESCANS, STRETCH_HALFTONE };
			for (int j = 0; j < 4; ++j)
			{
				[[maybe_unused]] int prev = SetStretchBltMode(hdc, modes[j]);
				for (int i = 0; i < 4; ++i)
				{
					StretchBlt(hdc, i * 100, 110 + j * 150, 100, 20 + i * 40, hdc, 0, 0, 100, 100, SRCCOPY);
				}
			}
			EndPaint(*this, &ps);
		}

		void on_keydown(USHORT vk_code, key_flags)
		{
			switch (vk_code)
			{
				case L'P':
					add_objects([] { return CreatePen(PS_SOLID, 0, RGB(255, 255, 0)); });
					return;
				case L'W':
					add_objects([] { return CreatePen(PS_SOLID, 1, RGB(255, 0, 255)); });
					return;
				case L'C':
					add_objects([] { 
						LOGBRUSH br{ .lbStyle = BS_SOLID, .lbColor = RGB(255, 0, 0) };
						return ExtCreatePen(PS_COSMETIC, 1, &br, 0, nullptr);
					});
					return;
				case L'G':
					add_objects([] {
						LOGBRUSH br{ .lbStyle = BS_SOLID, .lbColor = RGB(255, 0, 0) };
						return ExtCreatePen(PS_GEOMETRIC, 5, &br, 0, nullptr);
					});
					return;
				case L'B':
					add_objects([] { return CreateSolidBrush(RGB(0, 255, 255)); });
					return;
				case L'H':
					add_objects([] { return CreateHatchBrush(HS_DIAGCROSS, RGB(0, 255, 0)); });
				case L'L':
					m_table.list_objects();
					return;
				case L'Q':
					clear_objects();
					return;
					//r - rect region, e - elliptic region, d - device context, s - ddb, i - dib, p - palette
					//f - font, m - metafile, n - enhanced metafile, q - clear
			}
		}

		template<typename CreateFn>
		void add_objects(CreateFn &&fn)
		{
			int mult = 1;
			if (GetKeyState(VK_CONTROL) & 0x80)
				mult *= 10;
			if (GetKeyState(VK_SHIFT) & 0x80)
				mult *= 100;
			for (int i = 0; i < mult; ++i)
				m_objects.push_back(fn());
		}

		void clear_objects()
		{
			for (auto o : m_objects)
				DeleteObject(o);
			m_objects.clear();
		}

		std::vector<gdi_handle> m_objects;
		object_table m_table;
	};
}