module;

#include <windows.h>
#include <windowsx.h>

export module pige.paint;
export import pige.winapi;

export namespace pige::paint
{
	class paint_window : public winapi::window
	{
	public:
		using window::window;

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
			case WM_ERASEBKGND:
				return 0;
			case WM_PAINT:
			{
				PAINTSTRUCT ps;
				BeginPaint(*this, &ps);
				on_paint(ps.hdc);
				EndPaint(*this, &ps);
				return 0;
			}
			case WM_MOUSEMOVE:
				on_mousemove(GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam));
				return 0;
			case WM_LBUTTONDOWN:
				on_lbuttondown(GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam));
				return 0;
			case WM_RBUTTONDOWN:
				confirm_shape();
				return 0;
			case WM_KEYDOWN:
				on_keydown(static_cast<BYTE>(wparam));
				return 0;
			case WM_DESTROY:
				if (image)
					DeleteObject(image);
				break;
			}
			return window::window_proc(msg, wparam, lparam);
		}

	private:
		enum class Shape
		{
			None,
			Polyline = 'L',
			Polybezier = 'B',
			Arc = 'A',
			Rectangle = 'R',
			Ellipse = 'E',
			RoundRectangle = 'D',
			Chord = 'C',
			Pie = 'S',
			Polygon = 'P'
		};

		void invalidate_current()
		{
			if (pts.size() > 1)
			{
				auto p = pts.front();
				RECT rc{ p.x, p.y, p.x + 1, p.y + 1 };
				for (int i = 1; i < pts.size(); ++i)
				{
					p = pts[i];
					auto yp = std::minmax({ p.y, rc.top, rc.bottom });
					rc = RECT{ std::min(p.x, rc.left), std::min(p.y, rc.top), std::max(p.x + 1, rc.right), std::max(p.y + 1, rc.bottom) };
				}
				InvalidateRect(*this, &rc, true);
			}
		}

		void cancel_shape()
		{
			if (current == Shape::None) return;
			current = Shape::None;
			invalidate_current();
			pts.erase(pts.begin(), pts.end() - 1);
			SetWindowTextW(*this, L"Lines Beziers Arc Rectangle Ellipse rounDedrectangle Chord pieSlice Polygon");
		}

		void start_shape(Shape type)
		{
			cancel_shape();
			current = type;
			LPCWSTR title = L"Paint";
			switch (type)
			{
			case Shape::Polyline:
				title = L"Lines";
				break;
			case Shape::Polybezier:
				title = L"Beziers";
				break;
			case Shape::Arc:
				title = L"Arc";
				break;
			case Shape::Rectangle:
				title = L"Rectangle";
				break;
			case Shape::Ellipse:
				title = L"Ellipse";
				break;
			case Shape::RoundRectangle:
				title = L"Rounded Rectange";
				break;
			case Shape::Chord:
				title = L"Chord";
				break;
			case Shape::Pie:
				title = L"Pie Slice";
				break;
			case Shape::Polygon:
				title = L"Polygon";
				break;
			}
			SetWindowTextW(*this, title);
		}

		using PolyFn = decltype(Polyline);
		void draw_poly(HDC dc, PolyFn *fn)
		{
			fn(dc, pts.data(), static_cast<int>(pts.size()));
		}

		void draw_beziers(HDC dc)
		{
			int count = static_cast<int>((pts.size() - 1) / 3);
			count = 3 * count + 1;
			if (count < 1) return;
			PolyBezier(dc, pts.data(), count);
		}

		using ArcFn = decltype(Arc);
		void draw_arc(HDC dc, ArcFn *fn)
		{
			POINT p1 = pts[0], p2 = pts[1];
			POINT p3 = pts.size() > 2 ? pts[2] : p1;
			POINT p4 = pts.size() > 3 ? pts[3] : p3;
			fn(dc, p1.x, p1.y, p2.x, p2.y, p3.x, p3.y, p4.x, p4.y);
		}

		using RectFn = decltype(Rectangle);
		void draw_rect_impl(HDC dc, RectFn *fn)
		{
			fn(dc, pts[0].x, pts[0].y, pts[1].x, pts[1].y);
		}

		void draw_roundedrectangle(HDC dc)
		{
			POINT p1 = pts[0], p2 = pts[1];
			POINT p3 = pts.size() > 2 ? pts[2] : p1;
			int dx = std::min(std::abs(p3.x - p1.x), std::abs(p3.x - p2.x));
			int dy = std::min(std::abs(p3.y - p1.y), std::abs(p3.y - p2.y));
			RoundRect(dc, p1.x, p1.y, p2.x, p2.y, 2 * dx, 2 * dy);
		}

		void draw_shape(HDC dc)
		{
			if (pts.size() < 2) return;
			switch (current)
			{
			case Shape::None:
				return;
			case Shape::Polyline:
				draw_poly(dc, Polyline);
				return;
			case Shape::Polybezier:
				draw_beziers(dc);
				return;
			case Shape::Arc:
				draw_arc(dc, Arc);
				return;
			case Shape::Rectangle:
				draw_rect_impl(dc, Rectangle);
				return;
			case Shape::Ellipse:
				draw_rect_impl(dc, Ellipse);
				return;
			case Shape::RoundRectangle:
				draw_roundedrectangle(dc);
				return;
			case Shape::Chord:
				draw_arc(dc, Chord);
				return;
			case Shape::Pie:
				draw_arc(dc, Pie);
				return;
			case Shape::Polygon:
				draw_poly(dc, Polygon);
				return;
			}
		}

		void confirm_shape()
		{
			auto dc = GetDC(*this);
			auto mdc = CreateCompatibleDC(dc);
			auto oldBmp = SelectObject(mdc, image);
			invalidate_current();
			POINT last = pts.back();
			pts.erase(pts.end() - 1);
			draw_shape(mdc);
			pts.clear();
			pts.push_back(last);

			SelectObject(mdc, oldBmp);
			DeleteDC(mdc);
			ReleaseDC(*this, dc);
			cancel_shape();
		}

		void on_keydown(BYTE vk)
		{
			switch (vk)
			{
			case VK_ESCAPE:
				cancel_shape();
				return;
			case VK_RETURN:
				confirm_shape();
				return;
			case 'L':
			case 'B':
			case 'A':
			case 'R':
			case 'E':
			case 'D':
			case 'C':
			case 'S':
			case 'P':
				start_shape(static_cast<Shape>(vk));
				return;
			}
		}

		void on_mousemove(int x, int y)
		{
			pts.push_back({ x, y });
			invalidate_current();
			pts.erase(pts.end() - 2);
		}

		void on_lbuttondown(int x, int y)
		{
			if (current == Shape::None) return;

			POINT pt{ x,y };
			pts.back() = pt;
			if ((current == Shape::Arc || current == Shape::Chord || current == Shape::Pie) && pts.size() == 4 ||
				current == Shape::RoundRectangle && pts.size() == 3 ||
				(current == Shape::Rectangle || current == Shape::Ellipse) && pts.size() == 2)
			{
				pts.push_back(pt);
				confirm_shape();
				return;
			}
			pts.push_back(pt);
			invalidate_current();
		}

		void draw_background(HDC dc, RECT const &rc)
		{
			HDC mdc = CreateCompatibleDC(dc);
			HBITMAP oldBmp{};
			if (!image || image_size.cx != rc.right || image_size.cy != rc.bottom)
			{
				oldBmp = reinterpret_cast<HBITMAP>(SelectObject(mdc, CreateCompatibleBitmap(dc, rc.right, rc.bottom)));
				HBRUSH bgbr = reinterpret_cast<HBRUSH>(GetClassLongPtrW(*this, GCLP_HBRBACKGROUND));
				FillRect(mdc, &rc, bgbr);
				if (image)
				{
					auto idc = CreateCompatibleDC(dc);
					auto oldi = SelectObject(idc, image);
					BitBlt(mdc, 0, 0, image_size.cx, image_size.cy, idc, 0, 0, SRCCOPY);
					DeleteObject(SelectObject(idc, oldi));
					DeleteDC(idc);
				}
				image = reinterpret_cast<HBITMAP>(GetCurrentObject(mdc, OBJ_BITMAP));
				image_size = { rc.right, rc.bottom };
			}
			else
				oldBmp = reinterpret_cast<HBITMAP>(SelectObject(mdc, image));
			BitBlt(dc, 0, 0, rc.right, rc.bottom, mdc, 0, 0, SRCCOPY);
			SelectObject(mdc, oldBmp); //No delete here, we'd like to keep image alive
			DeleteDC(mdc);
		}

		void on_paint(HDC dc)
		{
			RECT rc{};
			GetClientRect(*this, &rc);
			if (current == Shape::None)
			{
				draw_background(dc, rc);
				return;
			}
			HDC mdc = CreateCompatibleDC(dc);
			auto oldBmp = SelectObject(mdc, CreateCompatibleBitmap(dc, rc.right, rc.bottom));
			draw_background(mdc, rc);
			draw_shape(mdc);
			
			BitBlt(dc, 0, 0, rc.right, rc.bottom, mdc, 0, 0, SRCCOPY);
			DeleteObject(SelectObject(mdc, oldBmp));
			DeleteDC(mdc);
		}

		HBITMAP image{};
		SIZE image_size{};
		Shape current = Shape::None;
		std::vector<POINT> pts{ { 0, 0} };
	};
}